import numpy as np
from mag.py.files import linesplit
def from_delim(file:str, delim:str=',', dtype=None, newline='\n') -> list:
    '''Parses deliminated file into numpy array.

    Args:
    '''
    with open(file, 'r') as f:
        data = np.array([linesplit(line, delim, newline) for line in f])

    if dtype is not None:
        data = data.astype(dtype)
    return data

def from_npz(
    filename:str,
    n_records:int=None,
    allow_pickle:bool=True,
    **kwargs
)->list:
    '''
    Notes:
        - NPZ loads data lazily, so it is efficient when just inspecting certain
            records, but less so when you want to map over all data.
    Arguments:
        filename (str): .npz file to load.
        n_records (int): Number of records to load. By default `None` which
            results in all data being loaded.
        allow_pickle (bool): whether or not to depickle saved objects.
        **kwargs: passed to `np.load`.
    Returns:
        results (list): list of all values from `filename`
    '''
    from sil import Sil
    print('NumPy lazy-loading index of saved data.')
    npz = np.load(filename, allow_pickle=allow_pickle, **kwargs)
    print('Loading data from NPZ file into memory. This may take a while.')
    files = npz.files if n_records is None else npz.files[:n_records]
    results = []
    status = Sil(len(files), every=100, estimate_time=True)
    for file in files:
        results.append(npz[file])
        status.tick()
    npz.close() # <-- prevent data leak
    return np.array(results)



import tensorflow as tf

class MultiLabelMacroRecall(tf.keras.metrics.Metric):
    def __init__(
        self, name='multi_label_macro_recall',
        threshold=0.5, from_logits=True, activation='sigmoid',
        **kwargs
    ):
        super(MultiLabelMacroRecall, self).__init__(name=name, **kwargs)
        self.threshold       = tf.constant(threshold)
        self.from_logits     = from_logits
        self.activation      = activation
        self.recall = self.add_weight(name='mlm_recall', initializer='zeros')
        self.true_positives  = self.add_weight(name='tp', initializer='zeros')
        self.false_positives = self.add_weight(name='fp', initializer='zeros')

    def update_state(self, y_true, y_pred):
        y_pred = tf.cond(
            tf.equal(self.from_logits, False), y_pred,
            tf.cond(
                tf.equal(self.activation, 'sigmoid'),
                tf.sigmoid(y_pred),
                y_pred
            )
        )

        # Compare predictions and threshold.
        pred_is_pos  = tf.greater(tf.cast(y_pred, tf.float32), self.threshold)
        label_is_pos = tf.greater(tf.cast(y_true, tf.float32), self.threshold)
        label_is_neg = tf.logical_not(tf.cast(label_is_pos, tf.bool))

        self.true_positives.assign_add(
            tf.reduce_sum(tf.cast(tf.logical_and(pred_is_pos, label_is_pos), tf.float32))
        )
        self.false_positives.assign_add(
            tf.reduce_sum(tf.cast(tf.logical_and(pred_is_pos, label_is_neg), tf.float32))
        )

        tp = self.true_negatives
        fp = self.false_positives
        recall = tf.div_no_nan(tp, tf.add(tp, fp))
        self.recall.assign(recall)
        return recall

    def result(self):
        return self.recall



class MultiLabelMacroSpecificity(tf.keras.metrics.Metric):

    def __init__(
        self, name='multi_label_macro_specificity',
        threshold=0.5, from_logits=True, activation='sigmoid',
        **kwargs
    ):
        super(MultiLabelMacroSpecificity, self).__init__(name=name, **kwargs)
        self.threshold       = tf.constant(threshold)
        self.from_logits     = from_logits
        self.activation      = activation

        self.specificity = self.add_weight(name='mlm_spec', initializer='zeros')
        self.true_negatives  = self.add_weight(name='tn', initializer='zeros')
        self.false_positives = self.add_weight(name='fp', initializer='zeros')

    def update_state(self, y_true, y_pred):
        y_pred = tf.cond(
            tf.equal(self.from_logits, False), y_pred,
            tf.cond(
                tf.equal(self.activation, 'sigmoid'),
                tf.sigmoid(y_pred),
                y_pred
            )
        )

        # Compare predictions and threshold.
        pred_is_pos  = tf.greater(tf.cast(y_pred, tf.float32), self.threshold)
        pred_is_neg  = tf.logical_not(tf.cast(pred_is_pos, tf.bool))
        label_is_pos = tf.greater(tf.cast(y_true, tf.float32), self.threshold)
        label_is_neg = tf.logical_not(tf.cast(label_is_pos, tf.bool))

        self.true_negatives.assign_add(
            tf.reduce_sum(tf.cast(tf.logical_and(pred_is_neg, label_is_neg), tf.float32))
        )
        self.false_positives.assign_add(
            tf.reduce_sum(tf.cast(tf.logical_and(pred_is_pos, label_is_neg), tf.float32))
        )

        tn = self.true_negatives
        fp = self.false_positives
        specificity = tf.div_no_nan(tn, tf.add(tn, fp))
        self.specificity.assign(specificity)
        return specificity

    def result(self):
        return self.specificity

class MultiLabelMacroSensitivity(tf.keras.metrics.Metric):

    def __init__(
        self, name='multi_label_macro_sensitivity',
        threshold=0.5, from_logits=True, activation='sigmoid',
        **kwargs
    ):
        super(MultiLabelMacroSensitivity, self).__init__(name=name, **kwargs)
        self.threshold       = tf.constant(threshold)
        self.from_logits     = from_logits
        self.activation      = activation

        self.sensitivity = self.add_weight(name='mlm_sens', initializer='zeros')
        self.true_positives  = self.add_weight(name='tp', initializer='zeros')
        self.false_negatives = self.add_weight(name='fn', initializer='zeros')

    def update_state(self, y_true, y_pred):
        y_pred = tf.cond(
            tf.equal(self.from_logits, False), y_pred,
            tf.cond(
                tf.equal(self.activation, 'sigmoid'),
                tf.sigmoid(y_pred),
                y_pred
            )
        )

        # Compare predictions and threshold.
        pred_is_pos  = tf.greater(tf.cast(y_pred, tf.float32), self.threshold)
        label_is_pos = tf.greater(tf.cast(y_true, tf.float32), self.threshold)

        pred_is_neg  = tf.logical_not(tf.cast(pred_is_pos, tf.bool))
        label_is_neg = tf.logical_not(tf.cast(label_is_pos, tf.bool))

        self.true_positives.assign_add(
            tf.reduce_sum(tf.cast(tf.logical_and(pred_is_pos, label_is_pos), tf.float32))
        )
        self.false_negatives.assign_add(
            tf.reduce_sum(tf.cast(tf.logical_and(pred_is_neg, label_is_pos), tf.float32))
        )

        tp = self.true_positives
        fn = self.false_negatives
        sensitivity = tf.div_no_nan(tp, tf.add(tp, fn))
        self.sensitivity.assign(sensitivity)
        return sensitivity

    def result(self):
        return self.sensitivity



import numpy as np
CONFUSION_METRIC_NAMES = [
    'True Positive Rate', 'True Negative Rate', 'Positive Predicitve Value',
    'Negative Predicitive Value', 'False Negative Rate', 'False Positive Rate',
    'False Discovery Rate', 'False Omission Rate', 'Threat Score', 'Accuracy',
    'F1 Score', 'Matthew\'s Correlation Coefficient', 'Informedness', 'Markedness'
]

def confusion_indicator_matrices(y_pred:list, y_true:list, threshold:float=0.5):
    '''
    Arguments:
        y_pred (list): the output matrix.
        y_true (list): the target matrix.
        threshold (float): values above `threshold` are `1`, rest are `0` for
            both `output` and `target`.
    Returns:
        tp (list): true postitives indicator matrix
        tn (list): true negatives indicator matrix
        fp (list): false positives indicator matrix
        fn (list): false negatives indicator matrix
    '''
    label_is_pos = np.greater(y_true, threshold)
    preds_is_pos = np.greater(y_pred, threshold)
    label_is_neg = np.logical_not(label_is_pos)
    preds_is_neg = np.logical_not(preds_is_pos)

    tp = np.logical_and(preds_is_pos, label_is_pos)
    tn = np.logical_and(preds_is_neg, label_is_neg)
    fp = np.logical_and(preds_is_pos, label_is_neg)
    fn = np.logical_and(preds_is_neg, label_is_pos)
    return tp, tn, fp, fn

def label_based_confusion_metrics(tp, tn, fp, fn):
    '''
    Notes:
        - expected arguments are output of `confusion_indicator_matrices`.

    Arguments:
        tp (list): true postitives indicator matrix
        tn (list): true negatives indicator matrix
        fp (list): false positives indicator matrix
        fn (list): false negatives indicator matrix

    Returns:
        label_confusion_metrics (np.ndarray): matrix with shape `[4, q]`, where
            `q` is the number of labels. The first 4 dimensions are the tp, tn,
            fp, and fn respectively.
    '''
    return np.array(list(map(lambda m: np.sum(m, axis=0), [tp, tn, fp, fn]))).T


def label_cardinality(labels):
    '''
    Arguments:
        labels (list): the 2d matrix of labels
    Returns:
        cardinality (float): average number of labels per example
    '''
    n_examples, n_labels = labels.shape
    return np.sum(labels) / n_examples

def label_density(labels, cardinality=None):
    '''
    Arguments:
        labels (list): the 2d matrix of labels
        cardinality (float): the `label_cardinality`. By default `None`. If
            `None` calculates the `label_cardinality`.
    Returns:
        density (float): normalized cardinality by possible number of labels
    '''
    n_examples, n_labels = labels.shape
    if cardinality is None:
        cardinality = label_cardinality(labels)
    return (1 / n_labels) * cardinality

def label_diversity(labels):
    '''
    Arguments:
        labels (list): the 2d matrix of labels
    Returns:
        diveristy (int): number of unique label sets that could occur
    '''
    n_examples, n_labels = labels.shape
    if n_examples > 10000:
        print('Calculating label diversity with {} examples.'.format(n_examples))
        print('This may take a while')

    return len(np.unique(labels, axis=0))

def proportional_label_diversity(labels, diversity=None):
    '''
    Arguments:
        labels (list): the 2d matrix of labels
        diveristy (int): The `label_diversity`. By default `None`. If `None`
            calculates the `label_diversity`.
    Returns:
        proportional_diversity (float): normalized diveristy by possible number
            examples
    '''
    n_examples, n_labels = labels.shape
    if diversity is None:
        diversity = label_diversity(labels)
    return (1 / n_examples) * diversity


DESCRIPTIVE_MULTI_LABEL_METRICS = {
    'label_cardinality': label_cardinality,
    'label_density': label_density,
    'label_diversity': label_diversity,
    'proportional_label_diversity': proportional_label_diversity
}
def descriptive_multi_label_metrics(
    y_true,
    threshold=0.5,
    metrics=list(DESCRIPTIVE_MULTI_LABEL_METRICS.keys())
):
    '''
    As defined in ["A review on Multi-Label Learning Algorithms"]
    (https://ieeexplore.ieee.org/stamp/stamp.jsp?tp=&arnumber=6471714&tag=1).

    Arguments:
        y_true (list): the target matrix.
        threshold (float): masks `y_true` (in the case of soft labeling).
        metrics (list): list of which metrics to compute. Choices include:
            label_cardinality: average number of labels per example.
            label_density: normalized cardinality by possible number of
                labels.
            label_diversity: number of unique label sets.
            proportional_label_diversity: normalized diversity by number of
                examples.
    Returns:
        results (dict): dictionary containing only the corresponding specified
            metrics, with types:
                label_cardinality (float)
                label_density (float)
                label_diversity (int)
                proportional_label_diversity (float)
    '''
    label_is_pos = np.greater(y_true, threshold)

    results = {}
    for metric in metrics:
        fn = DESCRIPTIVE_MULTI_LABEL_METRICS[metric]
        if fn is None: continue

        # save recomputation when possible
        if metric == 'label_density':
            if 'label_cardinality' in results:
                res = fn(label_is_pos, results['label_cardinality'])
            else:
                res = fn(label_is_pos)

        elif metric == 'proportional_label_diversity':
            if 'label_diversity' in results:
                res = fn(label_is_pos, results['label_diversity'])
            else:
                res = fn(label_is_pos)

        else:
            res = fn(label_is_pos)

        results[metric] = res
    return results


def true_positive_rate(tp, tn, fp, fn):
    '''
    Aliases:
        - recall
        - sensitivity
    Arguments:
        tp (int): number of true positives
        tn (int): number of true negatives
        fp (int): number of false positives
        fn (int): number of false negatives
    Returns:
        (float): [0, 1]
    '''
    if (tp + fn > 0):
        return tp / (tp + fn)
    if tp == 0:
        return 1 # found all the non existant hits
    return 0
recall = true_positive_rate
sensitivity = true_positive_rate

def true_negative_rate(tp, tn, fp, fn):
    '''
    Aliases:
        - specificity
    Arguments:
        tp (int): number of true positives
        tn (int): number of true negatives
        fp (int): number of false positives
        fn (int): number of false negatives
    Returns:
        (float): [0, 1]
    '''
    return tn / (tn + fp) if (tn + fp) > 0 else None
specificity = true_negative_rate

def positive_predictive_value(tp, tn, fp, fn):
    '''
    Aliases:
        - precision
    Arguments:
        tp (int): number of true positives
        tn (int): number of true negatives
        fp (int): number of false positives
        fn (int): number of false negatives
    Returns:
        (float): [0, 1]
    '''
    if (tp + fp) > 0:
        return tp / (tp + fp)
    if fp > 0:
        return 0
    return 1
precision = positive_predictive_value

def negative_predictive_value(tp, tn, fp, fn):
    '''
    Arguments:
        tp (int): number of true positives
        tn (int): number of true negatives
        fp (int): number of false positives
        fn (int): number of false negatives
    Returns:
        (float): [0, 1]
    '''
    return tn / (tn + fn) if (tn + fn) > 0 else None

def false_negative_rate(tp, tn, fp, fn):
    '''
    Arguments:
        tp (int): number of true positives
        tn (int): number of true negatives
        fp (int): number of false positives
        fn (int): number of false negatives
    Returns:
        (float): [0, 1]
    '''
    return fn / (fn + tp) if (fn + tp) > 0 else None

def false_positive_rate(tp, tn, fp, fn):
    '''
    Arguments:
        tp (int): number of true positives
        tn (int): number of true negatives
        fp (int): number of false positives
        fn (int): number of false negatives
    Returns:
        (float): [0, 1]
    '''
    return fp / (fp + tn) if (fp + tn) > 0 else None

def false_discovery_rate(tp, tn, fp, fn):
    '''
    Arguments:
        tp (int): number of true positives
        tn (int): number of true negatives
        fp (int): number of false positives
        fn (int): number of false negatives
    Returns:
        (float): [0, 1]
    '''
    return fp / (fp + tp) if (fp + tp) > 0 else None

def false_omission_rate(tp, tn, fp, fn):
    '''
    Arguments:
        tp (int): number of true positives
        tn (int): number of true negatives
        fp (int): number of false positives
        fn (int): number of false negatives
    Returns:
        (float): [0, 1]
    '''
    return fn / (fn + tn) if (fn + tn) > 0 else None

def threat_score(tp, tn, fp, fn):
    '''
    Arguments:
        tp (int): number of true positives
        tn (int): number of true negatives
        fp (int): number of false positives
        fn (int): number of false negatives
    Returns:
        (float): [0, 1]
    '''
    return tp / (tp + fn + fp) if (tp + fn + fp) > 0 else None

def accuracy(tp, tn, fp, fn):
    '''
    Arguments:
        tp (int): number of true positives
        tn (int): number of true negatives
        fp (int): number of false positives
        fn (int): number of false negatives
    Returns:
        (float): [0, 1]
    '''
    return (tp + tn) / (tp + tn + fp + fn) if (tp + tn + fp + fn) > 0 else None

def f1_score(tp, tn, fp, fn):
    '''
    Arguments:
        tp (int): number of true positives
        tn (int): number of true negatives
        fp (int): number of false positives
        fn (int): number of false negatives
    Returns:
        (float): [0, 1]
    '''
    return (2 * tp) / (2 * tp + fp + fn) if (2 * tp + fp + fn) > 0 else None

def matthews_correlation_coefficient(tp, tn, fp, fn):
    '''
    Arguments:
        tp (int): number of true positives
        tn (int): number of true negatives
        fp (int): number of false positives
        fn (int): number of false negatives
    Returns:
        (float): [0, 1]
    '''
    denominator = np.sqrt((tp + fp)*(tp + fn)*(tn + fp)*(tn + fn))
    return (tp * tn - fp * fn) / denominator if denominator > 0 else None

def bookmaker_informedness(tp, tn, fp, fn):
    '''
    Arguments:
        tp (int): number of true positives
        tn (int): number of true negatives
        fp (int): number of false positives
        fn (int): number of false negatives
    Returns:
        (float): [0, 1]
    '''
    sensitivity = true_positive_rate(tp, tn, fp, fn)
    specificity = true_negative_rate(tp, tn, fp, fn)
    valid_values = (sensitivity is not None and specificity is not None)
    return sensitivity + specificity - 1 if valid_values else None

def markedness(tp, tn, fp, fn):
    '''
    Arguments:
        tp (int): number of true positives
        tn (int): number of true negatives
        fp (int): number of false positives
        fn (int): number of false negatives
    Returns:
        (float): [0, 1]
    '''
    precision = positive_predictive_value(tp, tn, fp, fn)
    npv = negative_predictive_value(tp, tn, fp, fn)
    valid_values = (precision is not None and npv is not None)
    return precision + npv - 1 if valid_values else None

def micro_average_multi_label_metric(fn, lbcm:list):
    '''
    Arguments:
        fn (function): one of the `derived_confusion_metrics`, or a custom
            metric function with arguments signature `(tp, tn, fp, fn)`.
        lbcm (np.ndarray): The output of the function
            `label_based_confusion_metrics` or a matrix of shape `(4, q)`, where
            `q` is the number of label classes.
    Returns:
        result (float): the evaluated metric
    '''
    # sum over all label classes the [tp, tn, fn, fp]
    confusion_metrics = np.sum(lbcm, axis=0)
    try:
        result = fn(*confusion_metrics)
    except ZeroDivisionError:
        msg = (
            'Attempted to calculate the micro-average of {}'.format(fn.__name__)+
            'However, the denominator was 0. Returning `None`'
        )
        print(msg)
        return None
    return result

def individual_multi_label_metric(fn, lbcm:list):
    '''
    Arguments:
        fn (function): one of the `derived_confusion_metrics`, or a custom
            metric function with arguments signature `(tp, tn, fp, fn)`.
        lbcm (np.ndarray): The output of the function
            `label_based_confusion_metrics` or a matrix of shape `(4, q)`, where
            `q` is the number of label classes.
    Returns:
        result (list): the evaluated metric, per label
    '''
    res_per_label = list(map(lambda args: fn(*args), lbcm))
    return res_per_label

def macro_average_multi_label_metric(fn, lbcm:list):
    '''
    Arguments:
        fn (function): one of the `derived_confusion_metrics`, or a custom
            metric function with arguments signature `(tp, tn, fp, fn)`.
        lbcm (np.ndarray): The output of the function
            `label_based_confusion_metrics` or a matrix of shape `(4, q)`, where
            `q` is the number of label classes.
    Returns:
        result (float): the evaluated metric
    '''
    # average metric by number of label classes
    div = 1 / lbcm.shape[0]
    try:
        res_per_label = individual_multi_label_metric(fn, lbcm)
        results = div * np.sum(res_per_label)
    except TypeError:
        nones = np.where(([1, None], None))[0].tolist()
        msg = (
            'Label classes at indicies {} are None.'.format(nones)+
            'Can not sum with unevaluateable metrics. Returning `None`.'
        )
        if nones:
            print(msg)
        return None
    return results

DERIVED_MULTI_LABEL_METRICS = {
    'true_positive_rate': true_positive_rate,
    'recall': true_positive_rate,
    'sensitivity': true_positive_rate,

    'true_negative_rate': true_negative_rate,
    'specificity': true_negative_rate,

    'positive_predictive_value': positive_predictive_value,
    'precision': positive_predictive_value,

    'negative_predictive_value': negative_predictive_value,
    'false_negative_rate': false_negative_rate,
    'false_positive_rate': false_positive_rate,
    'false_discovery_rate': false_discovery_rate,
    'false_omission_rate': false_omission_rate,
    'threat_score': threat_score,
    'accuracy': accuracy,
    'f1_score': f1_score,
    'matthews_correlation_coefficient': matthews_correlation_coefficient,
    'bookmaker_informedness': bookmaker_informedness,
    'markedness': markedness
}
def derived_multi_label_metrics(
    label_based_confusion_matrix:list,
    metrics=['accuracy', 'precision', 'recall', 'f1_score'],
    method='macro'
):
    '''
    Arguments:
        label_based_confusion_matrix (np.ndarray): a matrix with shape `(4, q)`
            where `q` is the number of label classes and the first dimension
            corresponds to tp, tn, fp, and fn respectively.

        metrics (list): list of metrics to calculate. By default,
            `['accuracy', 'precision', 'recall', 'f1_score']` Options include:
                true_positive_rate (alisases: recall, sensitivity)
                true_negative_rate (alias: specificity)
                positive_predictive_value (alias: precision)
                negative_predictive_value
                false_negative_rate
                false_positive_rate
                false_discovery_rate
                false_omission_rate
                threat_score
                accuracy
                f1_score
                matthews_correlation_coefficient
                sensitivity
                markedness

        method (str): how to apply the metrics specified, choices include:
            micro, macro and individual. Defaults to `macro`.

    Returns:
        results (dict): the specified metrics and corresponding value.
    '''
    if method not in {'macro', 'micro', 'individual'}:
        method = 'micro'

    if method == 'micro':
        eval_fn = micro_average_multi_label_metric
    elif method == 'individual':
        eval_fn = individual_multi_label_metric
    else:
        eval_fn = macro_average_multi_label_metric

    results = {}
    for metric in metrics:
        fn = DERIVED_MULTI_LABEL_METRICS[metric]
        if fn == None: continue
        try:
            res = eval_fn(fn, label_based_confusion_matrix)
        except Exception as e:
            print(e)
            res = None
        results['{}'.format(metric)] = res
    return results

def multi_label_metric_report(
    label_based_confusion_matrix:list,
    metrics:list=['accuracy', 'precision', 'recall', 'f1_score'],
    label_names:list=None,
    verbose:bool=False
):
    '''
    Arguments:
        label_based_confusion_matrix (np.ndarray): a matrix with shape `(4, q)`
            where `q` is the number of label classes and the first dimension
            corresponds to tp, tn, fp, and fn respectively.

        metrics (list): list of metrics to calculate. By default,
            `['accuracy', 'precision', 'recall', 'f1_score']` Options include:
                true_positive_rate (alisases: recall, sensitivity)
                true_negative_rate (alias: specificity)
                positive_predictive_value (alias: precision)
                negative_predictive_value
                false_negative_rate
                false_positive_rate
                false_discovery_rate
                false_omission_rate
                threat_score
                accuracy
                f1_score
                matthews_correlation_coefficient
                sensitivity
                markedness

        label_names (list): list of the label classes, used for printing the
            metrics report.

        verbose (bool): whether or not to print a formatted string of the
            results.

    Returns:
        results (dict): consists of three main keys, `macro`, `micro`, and
            `individual`, and their subdictionaries consiting of the specified
            metrics.
    '''
    macros = derived_multi_label_metrics(
        label_based_confusion_matrix,
        method='macro',
        metrics=metrics
    )

    micros = derived_multi_label_metrics(
        label_based_confusion_matrix,
        method='micro',
        metrics=metrics
    )

    indivs = derived_multi_label_metrics(
        label_based_confusion_matrix,
        method='individual',
        metrics=metrics
    )

    if label_names == None:
        names = ['Label {}'.format(i) for i in range(lbcm.shape[0])]
    else:
        names = label_names

    outstring = 'Macros:\n------------------------------------------------\n'
    longest_metric_name = max(list(map(len, list(macros.keys()))))
    for k, v in macros.items():
        spaces = ' ' * (4 + longest_metric_name-len(k))
        outstring += '{}:{}{}\n'.format(k, spaces, np.round(v, 3))

    outstring += '\nMicros:\n------------------------------------------------\n'
    longest_metric_name = max(list(map(len, list(macros.keys()))))
    for k, v in micros.items():
        spaces = ' ' * (4 + longest_metric_name-len(k))
        outstring += '{}:{}{}\n'.format(k, spaces, np.round(v, 3))

    outstring += '\nPer Label:\n------------------------------------------------\n'


    longest_metric_name  = max(list(map(len, list(indivs.keys()))))
    longest_name = max(list(map(len, names)))
    for i, name in enumerate(names):
        outstring += '{}:\n'.format(name)
        for k, v in indivs.items():
            longest = (longest_metric_name+longest_name)
            spaces = ' ' * (4 + longest - len(k) - len(name))
            outstring += '\t{}:{}{}\n'.format(k, spaces, np.round(v[i], 3))

    if verbose:
        print(outstring)

    return {'macro': macros, 'micro': micros, 'individual': indivs}



import numpy as np
def unravel_merged_matrices(merged_flattened_index:int, elements_in_matrices:list):
    '''
    Arguments:
        merged_flattened_index (int): the index of the element in the merged,
            flattened matrix
        elements_in_matrices (list): a list of integers corresponding to the
            number of elements in the matrix at the corresponding index
    Returns:
        flattened_index (int): the flattened index for the matrix to which
            merged_flattened_index belongs to
        matrix_index (int): the index correspoinding to the matrix to which the
            flattened_index belongs to

    '''
    flattened_index = merged_flattened_index
    matrix_index = -1
    for i, num_elements in enumerate(elements_in_matrices):
        if flattened_index < num_elements:
            matrix_index = i
            break
        flattened_index -= num_elements
    return (flattened_index, matrix_index)

def merge_matrices(matrices:list):
    '''
    Arguments:
        matrices (list): a list of matrices (np.ndarray) to merge
    Returns:
        merged (list): a 1D np.ndarray consist of all the elements from matrices
        flattened_matrices (list): a list of 1D np.ndarrays corresponding to the
            flattened matrices from matrices
        elements_per_matrix (list): the number of elements per matrix in
            matrices
    '''
    # dealing with matrices of different shape is cumbersome, so flatten them
    # and leverage 1D mapping.
    flattened_matrices = [mat.flatten() for mat in matrices]

    # save some re compuation
    elements_per_matrix = list(map(len, flattened_matrices))

    # one long, flattened matrix
    merged = np.concatenate(flattened_matrices)
    return merged, flattened_matrices, elements_per_matrix

def matrix_indices_from_merged_array(
    elements_needed:int,
    merged_array:list,
    elements_per_array:list
):
    '''
    Arguments:
        elements_needed (int): how many elements to retrieve

        merged_array (list): a 1D numpy.ndarray that consists of several
            distinct n-dimensional matrices which were flattened and
            concatenate and joined together.

        elements_per_array (list): a list of the total number of elements in
            each of the n-dimensional matrices which make up the merged_array

    Returns:
        raveled_index_matrix_tuples (list): a list of tuples which consists of
            raveled_index (int): the raveled (flattened) index of the element
            matrix_index (int): the matrix to which the raveled_index belongs
                from which merged_array is made from.
    '''
    # argpartition is faster than argmin, but the top smallest k values are not
    # sorted
    smallest_elements = np.argpartition(merged_array, elements_needed)[:elements_needed]
    raveled_index_matrix_tuples = list(map(
        lambda i: unravel_merged_matrices(i, elements_per_array),
        smallest_elements
    ))
    return raveled_index_matrix_tuples

def raveled_indices_of_matrix(raveled_index_matrix_tuples:list, matrix_index:int):
    '''
    Arguments:
        raveled_index_matrix_tuples (list): a list of (raveled_index:int,
            matrix_index:int) tuples
        matrix_index (int): the matrix from which raveled_indices are requested
    Returns:
        indices (list): a list of raveled indices belonging to matrix_index
    '''
    belongs_to_matrix_q = lambda tup: tup[-1] == matrix_index
    index_extractor = lambda tup: tup[0]
    indices = list(map(index_extractor, filter(belongs_to_matrix_q, raveled_index_matrix_tuples)))
    return indices




def shard_rng(maxlen, sublen, stride):
    '''
    A wrapper around `range(start, stop, step)` such that if stop is not evenly
    divisible by step, the last step is `(stop-sublen, stop)`.

    Arguments:
        maxlen (int): the length to shard into steps.
        sublen (int): how long each substep will be.
        stride (int): how far to step.

    Returns:
        ranges (list): a list of (start, stop) sublists. These ranges may overlap.

    '''
    ranges = []
    for i in range(0, maxlen, stride):
        if i + sublen > maxlen:
            ranges.append([maxlen-sublen, maxlen])
            break
        else:
            ranges.append([i, i+sublen])
    return ranges

def weight_rngs(ranges):
    '''
    Helper function for `stitch_mats` with `method=1` (`_stitch1`).
    Arguments:
        ranges (list): a list of (start, stop) sublists.

    Returns:
        weights (list): a 1D array of length `ranges[-1][-1]`. How much to scale
            each index due to overlapping steps.
    '''
    n = ranges[-1][-1]
    bins = map(np.bincount,np.array(ranges).T,(None,None),(n+1,n+1))
    vals = np.subtract(*bins).cumsum()
    weights = 1 / vals[:n,None]
    return weights


def _stitch1(shape, submats, ranges):
    stitched = np.zeros(shape)
    weights = weight_rngs(ranges)
    for submat, (start, stop) in zip(submats, ranges):
        stitched[start:stop] += weights[start:stop] * submat
    return stitched


def _stitch2(shape, submats, ranges):
    ranges = np.array(ranges)
    ro = ranges.ravel().argsort(kind='stable')

    # put 1 for starting and -1 for ending, take cumsum
    cnts = (1-((ro&1)<<1)).cumsum()

    stitched = np.zeros((n,m))
    # add slices
    for submat, (start, stop) in zip(submats,ranges):
        stitched[start:stop] += submat

    rs = ranges.ravel()[ro]
    # divide by overlap
    for start, stop, count in zip(rs[:-1],rs[1:],cnts[:-1]):
        stitched[start:stop] /= count
    return stitched


def stitch_mats(shape, submats, ranges, method=1):
    '''
    Arguments:
        shape (tuple): the shape of the output matrix `(n, m)`.
        submats (list): a list of matrices all with shape `(k, m)`.
        ranges (list): a list of (start, stop) sublists "ranges". These ranges
            are across the first dimension of the output matrix and specifiy
            which region of `n` each submatrix corresponds to.
        method (int): either `1` or `2`. Returns identically results, may affect
            runtime.

    Returns:
        stitched (list): a matrix of shape `(n, m)` where the submats are joined
            together, averaging the values when they overlap (based on `ranges`).
    '''
    fn = _stitch1 if method == 1 else _stitch2
    return fn(shape, submats, ranges)

def shard_mat(mat, ranges):
    '''
    Arguments:
        mat (list): a matrix of shape `(n, m)` to produce submatrices from.
        ranges (list): a list of (start, stop) sublists indicating how to make
            the submatricies (e.g. `mat[start:stop]`)
    Returns:
        submats (list): a list of submats that start and stop as specified in
            `ranges`.
    '''
    return [mat[start:stop] for start, stop in ranges]



from tensorflow.python.client import device_lib
import os, tensorflow as tf

def available_devices():
    '''

    Returns:
        devices(dict): a list of devices by type (cpu / gpu)
    '''
    devices = device_lib.list_local_devices()
    cpus = [d for d in devices if d.device_type == 'CPU']
    gpus = [d for d in devices if d.device_type == 'GPU']
    return {'cpus': cpus, 'gpus': gpus}


def gpu_available_q():
    '''
    Returns:
        (bool): whether or not there is an availble gpu
    '''
    return len(available_devices()['gpus']) > 0


def set_env_devices(devices:list):
    '''
    Sets the environment variable CUDA_VISIBLE_DEVICES to devices

    Args:
        devices(list): a list of device names

    Returns:
        None

    '''
    os.environ["CUDA_VISIBLE_DEVICES"] = ','.join(devices)

def restrict_gpu_devices(
    device_names:list=None,
    num_devices:int=None,
    verbose=False
):
    '''
    Utilizes experimental tf functions to restrict GPU devices. Must be called
    prior to other tf functions.

    Notes:
        - Only either `device_names` or `num_devices` can be set. `num_devices`
            takes precedent over `device_names`

    Arguments:
        device_names (list): Name of devices to use. Defaults to None. If None,
            uses all available devices

        num_devices (int): Number of devices to use. Defaults to None. Takes
            _first_ `num_devices` GPUs.

        verbose (bool): whether or not to print out devices. Defaults to False.

    Returns:
        None
    '''
    env_gpus = tf.config.experimental.list_physical_devices('GPU')
    if verbose:
        print('GPUs found in environment:\t', env_gpus)

    if device_names is not None:
        gpus = [g for g in env_gpus if g.name in device_names]

    if num_devices is not None:
        gpus = env_gpus[:num_devices]

    if verbose:
        print('GPUs found matching device names:\t', gpus)
    if gpus:
        try:
            tf.config.experimental.set_visible_devices(gpus, 'GPU')
            logical_gpus = tf.config.experimental.list_logical_devices('GPU')
            if verbose:
                print('Environment has {} physical GPUs'.format(env_gpus))
                print('Restricted to {} logical GPUs'.format(logical_gpus))
        except RuntimeError as e:
            # Visible devices must be set before GPUs have been initialized
            print(e)
    else:
        if verbose:
            print('No GPUs found.')



import numpy as np

def binarize(data:list, cutoff:float=0.5) -> list:
    '''Converts all values of data to either 1 or 0 depending on
    whether or not the value is greater than cutoff.

    Args:
        data (ndarray): an arbitrary array of numerical values
        cutoff (float): a value bounded [0, 1] determining which values are
            converted to one.

    Returns:
        binary_data (ndarray:int): an array with same dimensions as data where
            all values are either 0 or 1
    '''
    return (data > cutoff).astype(int)

def consecutive(data:list, stepsize:float=1) -> list:
    '''Splits data into subarrays where the values in each array vary by stepsize

    Args:
        data (ndarray): a one dimensional array of sorted values
        stepsize (float): a numeric value specifying how close values should be
            to be considered consecutive
    Returns:
        consecutive_numbers (ndarray): a list of values from data where each
            sublist contains values that are exactly stepsize apart
    '''
    return np.split(data, np.where(np.diff(data) != stepsize)[0]+1)

from mag.py.math import domain
def runs(array:list) -> list:
    '''Returns the [start, stop] values of consectuive numbers in the given array

    Args:
        array (list): a list of values, sorted least to greatest

    Returns:
        runs (list): a list of all runs found in the array, where a run specifies
            the lower and upper bounds of consecutive numbers.
    '''
    return [domain(run) if run.size > 0 else [] for run in consecutive(array)]


def nonzero_1d(data:list) -> list:
    '''Returns the indices of nonzero values in an one-dimensional array.

    Args:
        data (ndarray): an array with one dimension

    Returns:
        indices (ndarray): a one dimensional array containing indicies of data
            which are nonzero
    '''
    return data.nonzero()[0]

def are_same_run(run_a:list, run_b:list, tolerance:int=0) -> bool:
    '''
    Notes:
        - Used by `smooth_runs`
    Arguments:
        run_a (list): a list of intergers consisting of `[start, stop]`
        run_b (list): a list of intergers consisting of `[start, stop]`
        tolerance (int): how far apart run_a and run_b to be considered
            a part of the same run. By default `0`.
    Returns:
        are_same_run (bool): whether or not the two runs are within the
            specified `tolerance`.
    '''
    start_a, stop_a = run_a
    start_b, stop_b = run_b
    if start_a < start_b:
        return start_b - stop_a <= tolerance + 1
    return start_a - stop_b <= tolerance + 1

def merge_runs(run_a:list, run_b:list) -> list:
    return [min(run_a[0], run_b[0]), max(run_a[1], run_b[1])]



def smooth_runs(runs:list, tolerance:int=0):
    '''
    Notes:
        - Assumes `runs` is a list of lists, where each sublist is specifies a
            "run" (start and stop indices) for a given feature. Expected to be
            the output of the function `runs`.

        - Assumes `runs` is _ordered_ (e.g. `runs[0][0] < runs[1][0]`)

    Arguments:
        runs (list): a list of runs, where a run specifies the lower and upper
        bounds of a range.

        tolerance (int): How many _indicies_ between two runs in `runs` for them
            to be considered part of the same run. By default `tolerance` is
            set to `0`.

    Returns:
        smoothed (list): a lists of lists (similar to `runs`), where each run is
            seperated by at most `tolerance` indicies are merged.
    '''
    num_runs = np.inf
    smoothed = np.array(runs)
    while num_runs > len(smoothed):
        # update stopping condition
        num_runs = len(smoothed)

        # collection var for merged runs
        coalesced = np.array([0][:0]).reshape(-1,2)
        for i in range(len(smoothed)):
            run_a = smoothed[i]
            append_flag = True # not yet in coalesced
            for j in range(len(coalesced)):
                run_b = coalesced[j]
                if are_same_run(run_a, run_b, tolerance):
                    append_flag = False # merge into exisitng run
                    coalesced[j] = merge_runs(run_a, run_b)
                    break
            if append_flag:
                coalesced = np.concatenate((coalesced, np.array([run_a])))
        # update smoothed
        smoothed = coalesced
    return smoothed.astype(int).tolist()
